/*
**  Cubes.c    Double buffered 3D animation demo
**
**  6/21/07 v2.0 Lucio Di Jasio
**  8/03/07 v2.1 LDJ automated generation of vertexes
**
**  HW Req: Explorer16 board + AV1632 with NTSC/PAL output
**  SW Req: Explore.c Geometry2.c Graphic.c
**  Ref:    www.exploringpic32.com
** 
**  This demo uses integer math to perform 3D drawing and 
**  animation with Isometric perspective
**  There are truly 3 demos, depending on which button
**  is pressed on the Explorer16 board after start up:
**  1- leftmosts button shows a single cube spinning
**  2- second from left, shows 38 cubes forming the digits 32
**  3- third button from left, show 100+ cubes forming the 
**      string: "I (love) PIC32" 
**      - press once more (any) button to start (scramble)
**      - press and keep pressed (any) button to have message
**          gravitate back in place
*/


// configuration bit settings
#pragma config POSCMOD=XT, FNOSC=PRIPLL
#pragma config FPLLIDIV=DIV_2, FPLLMUL=MUL_18, FPLLODIV=DIV_1
#pragma config FWDTEN=OFF, CP=OFF, BWP=OFF

#include <p32xxxx.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <Explore.h>
#include <Graphic.h>
#include "Geometry2.h"


extern const char Font8x8[];
extern short VCount;

// define the animation step (angle)
#define STEP    5  

// define the points table
#define MAXP    1100
point plist[ MAXP];           
short pc = MAXP;                    // points counter

// define the vectors table
#define MAXV    1600 
vector vlist[ MAXV];
short vc = MAXV;                    // vectors counter

// define the polygons table
#define MAXPOLY 150
poly pylist[ MAXPOLY];
short pyc = MAXPOLY;                  // polygons counter

// define the objects table
#define MAXOBJ 10
object olist[ MAXOBJ];
short objc = MAXOBJ;                  // object counter



void drawPolygons( short p)
{   // draws all vectors in the poligon of index p
    short i, x;
    point a, b, c;
    
    x = pylist[ p].v;
    for ( i=0; i<pylist[p].n; i++)
    { 
        // compute the first point coordinates 
        matrixProd( plist[vlist[x][0]], a);
        normalize( a);
        perspective( a);

        // compute the second point coordinates 
        matrixProd( plist[vlist[x++][1]], b);
        normalize( b);
        perspective( b);

        // draw the vector
        line( a[0], a[1], b[0], b[1] );

    } // for
} // drawPolygons

void drawObject( short obj)
{
    short i;
    
    for( i=0; i< olist[ obj].n; i++)
        drawPolygons( olist[ obj].p+i);

}// drawObject



short appPoint( short x, short y, short z)
{ // add a point to the list and return its index
    point *p = &plist[--pc];

    (*p)[0] = x; (*p)[1] = y; (*p)[2] = z;
    
    return pc;
} // appPoint


short cube( short x, short y, short z, short s)
{ // generates a cube at x,y,z coordinates, of side s
  // represented as a single poly (optimized for speed, min vectors) 
  // cannot be used for surface removal!!!

  short a, b ,c, d, e, f, g, h;
  // verify there is room in the tables
    if ((pyc>0) && (vc>11) && (pc>7))
    {
      // generate all vertexes of the cube
        a = appPoint( x, y, z);        // A
        b = appPoint( x, y+s, z);      // B
        c = appPoint( x+s, y+s, z);    // C
        d = appPoint( x+s, y, z);      // D
        e = appPoint( x+s, y, z+s);    // E
        f = appPoint( x+s, y+s, z+s);  // F
        g = appPoint( x, y+s, z+s);    // G
        h = appPoint( x, y, z+s);      // H

      // generate all the vectors
        vlist[ --vc][0] = a; vlist[ vc][1] = b;
        vlist[ --vc][0] = b; vlist[ vc][1] = c;
        vlist[ --vc][0] = c; vlist[ vc][1] = d;
        vlist[ --vc][0] = d; vlist[ vc][1] = a;
        vlist[ --vc][0] = e; vlist[ vc][1] = f;
        vlist[ --vc][0] = f; vlist[ vc][1] = g;
        vlist[ --vc][0] = g; vlist[ vc][1] = h;
        vlist[ --vc][0] = h; vlist[ vc][1] = e;
        vlist[ --vc][0] = b; vlist[ vc][1] = g;
        vlist[ --vc][0] = a; vlist[ vc][1] = h;
        vlist[ --vc][0] = c; vlist[ vc][1] = f;
        vlist[ --vc][0] = d; vlist[ vc][1] = e;

      // generate a single poligon structure containing all the vectors
        pylist[ --pyc].v = vc;          // the last vector created has the highest index
        pylist[ pyc].n = 12;            // 12 vectors

      // returns the polygon index
        return pyc;
    } // if 
    else 
        while( 1);
} // cube


void createObject( int m, char *s, int w)
{  // creates a pattern of cubes from a (short) text string
    // w is the desired total string width
    // m mode: 0- single cube
    //         1- multiple objects
    char c;
    const char*p;
    int i, j, wc, x0, x, y;
    short py, n;

    // determine maximum allowed size of each dot/cube
    wc = w /(8*strlen( s));

    // determine initial horizontal position
    x0 = - w/2;

    // start counting the cubes
    n = 0;

    // map each character in the string
    while(*s)
    { // determine intial position
        y = + 4*wc;
        // find character in font
        p = &Font8x8[(*s-0x20)<<3];
        for(i=0; i<8; i++)
        {   // one line at a time
            c = *p++;
            x = x0;
            for(j=0; j<8; j++)
            { // one pixel at a time
                if ( c & 0x80)
                {
                    py = cube( x, y, 0, wc);
                    n++;
                }
                x+=wc;
                c<<=1;
            } // forj
            y -= wc;
        } // for i
        s++;
        x0 += 8 * wc;

        if ( m)
        { // generate an object for each character in the string
          olist[ --objc].p = py;  // the last cube/polygon inserted 
          olist[ objc].n = n;     // the number of cubes created
          n = 0;                  // reset counter of cubes
        }
    } // while

    if ( !m)
    { // generate a single object for all characters
      olist[ --objc].p = py;  // the last cube/polygon inserted 
      olist[ objc].n = n;     // the number of cubes created
    }            
} // creatObject     
   

short x0, y0, z0;
short a[MAXOBJ], b[MAXOBJ], g[MAXOBJ];
short r[MAXOBJ], k[MAXOBJ];


int main( void)
{
    short v, i, j;
    char start=0;
    char s[32];
 
    // init angles and offsets
    x0 = y0 = 130;
    z0 = 100;
    
    // init all objects angles
    for( i=objc; i<MAXOBJ; i++)
        a[i] = b[i] = g[i] = 0;

    objc = MAXOBJ;
    vc = MAXV;
    pyc = MAXPOLY;
    pc = MAXP;


    // hardware and video initialization 
    TRISA = 0xfff0;
    T1CON = 0x8030;
    initEX16();                  

    initVideo();
    clearScreen();                 

    while( !readKEY());
    srand( TMR1);
    
    if( getKEY()== 8)              // first button
        createObject( 0, "\\", 800); // single cube
    else if( getKEY() == 4)        // second button 32
        createObject( 0, "32", 350);                 
    else 
    {                          // third button cube
        createObject( 1, "I`PIC32", 350);
        start = 1;
    }

    // show the totals
    sprintf( s, "%d Vrtx  %d Seg  %d Obj", MAXP-pc, MAXV-vc, MAXPOLY-pyc);
        
    // 3. main loop
    while( 1)
    {
        for( i=objc; i<MAXOBJ; i++)
        {
            // pick random movements
            r[i] = ( rand() & 1) ? STEP : -STEP;  // random verse
            k[i] = ( rand() % 3);           // random axis
        }
        j = 5 + (rand() & 0x3f);            // random rotation

        // animate
        while ( j-- > 0)
        {  
            clearHScreen();                 // clear the hidden screen
            AT( 1, 1); putsV( s);

            for( i=objc; i<MAXOBJ; i++)
            {
                initWorld( a[i], b[i], g[i], x0, y0, z0);
                drawObject( i);
                if ( readKEY())
                { // return all objects to base position
                    if ( a[i] >0 )
                        a[i]-= STEP;
                    else if ( a[i] <0)
                        a[i]+= STEP;
                    if ( b[i] >0 )
                        b[i]-= STEP;
                    else if ( b[i] <0)
                        b[i]+= STEP;
                    if ( g[i] >0 )
                        g[i]-= STEP;
                    else if ( g[i] <0)
                        g[i]+= STEP;
                    
                }
                else 
                //perform the update
                switch ( k[i]){
                case 0:
                    a[i] += r[i];
                    break;
                case 1: 
                    b[i] += r[i];
                    break;
                case 2:
                    g[i] += r[i];
                    break;
                } // switch
            } // for

            // show the first position
            swapV();
            
            // wait for button to start the animation
            if ( !start)
                start = getKEY();
            //else
            
        } // while animation           


    } // main loop
} // main
